"""Morgan Stanley parser.

Note, that I only had access to an Alphabet export. I have no idea how it looks like
for another company, or for a full profile.
"""
from __future__ import annotations

import csv
from datetime import datetime
from decimal import Decimal
from pathlib import Path
from typing import Final

from cgt_calc.exceptions import ParsingError, UnexpectedColumnCountError
from cgt_calc.model import ActionType, BrokerTransaction

COLUMNS_RELEASE: Final[list[str]] = [
    "Vest Date",
    "Order Number",
    "Plan",
    "Type",
    "Status",
    "Price",
    "Quantity",
    "Net Cash Proceeds",
    "Net Share Proceeds",
    "Tax Payment Method",
]

COLUMNS_WITHDRAWAL: Final[list[str]] = [
    "Date",
    "Order Number",
    "Plan",
    "Type",
    "Order Status",
    "Price",
    "Quantity",
    "Net Amount",
    "Net Share Proceeds",
    "Tax Payment Method",
]

# These can be potentially wired through as a flag
KNOWN_SYMBOL_DICT: Final[dict[str, str]] = {
    "GSU Class C": "GOOG",
}


def _hacky_parse_decimal(decimal: str) -> Decimal:
    return Decimal(decimal.replace(",", ""))


def _init_from_release_report(row_raw: list[str], filename: str) -> BrokerTransaction:
    if len(COLUMNS_RELEASE) != len(row_raw):
        raise UnexpectedColumnCountError(row_raw, len(COLUMNS_RELEASE), filename)
    row = {col: row_raw[i] for i, col in enumerate(COLUMNS_RELEASE)}

    if row["Type"] != "Release":
        raise ParsingError(filename, f"Unknown type: {row_raw[3]}")

    if row["Status"] != "Complete":
        raise ParsingError(filename, f"Unknown status: {row_raw[5]}")

    if row["Price"][0] != "$":
        raise ParsingError(filename, f"Unknown price currency: {row_raw[6]}")

    if row["Net Cash Proceeds"] != "$0.00":
        raise ParsingError(filename, f"Non-zero Net Cash Proceeds: {row_raw[8]}")

    if row["Plan"] not in KNOWN_SYMBOL_DICT:
        raise ParsingError(filename, f"Unknown plan: {row_raw[3]}")

    quantity = _hacky_parse_decimal(row["Net Share Proceeds"])
    price = _hacky_parse_decimal(row["Price"][1:])
    amount = quantity * price

    return BrokerTransaction(
        date=datetime.strptime(row["Vest Date"], "%d-%b-%Y").date(),
        action=ActionType.STOCK_ACTIVITY,
        symbol=KNOWN_SYMBOL_DICT[row["Plan"]],
        description=row["Plan"],
        quantity=quantity,
        price=price,
        fees=Decimal(0),
        amount=amount,
        currency="USD",
        broker="Morgan Stanley",
    )


def _init_from_withdrawal_report(
    row_raw: list[str], filename: str
) -> BrokerTransaction:
    if len(COLUMNS_WITHDRAWAL) != len(row_raw):
        raise UnexpectedColumnCountError(row_raw, len(COLUMNS_WITHDRAWAL), filename)
    row = {col: row_raw[i] for i, col in enumerate(COLUMNS_WITHDRAWAL)}

    if row["Type"] != "Sale":
        raise ParsingError(filename, f"Unknown type: {row_raw[3]}")

    if row["Order Status"] != "Complete":
        raise ParsingError(filename, f"Unknown status: {row_raw[5]}")

    if row["Price"][0] != "$":
        raise ParsingError(filename, f"Unknown price currency: {row_raw[6]}")

    if row["Plan"] not in KNOWN_SYMBOL_DICT:
        raise ParsingError(filename, f"Unknown plan: {row_raw[3]}")

    quantity = -Decimal(row["Quantity"])
    price = _hacky_parse_decimal(row["Price"][1:])
    amount = _hacky_parse_decimal(row["Net Amount"][1:])

    return BrokerTransaction(
        date=datetime.strptime(row["Date"], "%d-%b-%Y").date(),
        action=ActionType.SELL,
        symbol=KNOWN_SYMBOL_DICT[row["Plan"]],
        description=row["Plan"],
        quantity=quantity,
        price=price,
        fees=quantity * price - amount,
        amount=amount,
        currency="USD",
        broker="Morgan Stanley",
    )


def _validate_header(
    header: list[str], golden_header: list[str], filename: str
) -> None:
    """Check if header is valid."""
    if len(golden_header) != len(header):
        raise UnexpectedColumnCountError(header, len(golden_header), filename)
    for i, (expected, actual) in enumerate(zip(golden_header, header)):
        if expected != actual:
            msg = f"Expected column {i+1} to be {expected} but found {actual}"
            raise ParsingError(filename, msg)


def read_mssb_transactions(transactions_folder: str) -> list[BrokerTransaction]:
    """Parse Morgan Stanley transactions from CSV file."""
    transactions = []

    for file in Path(transactions_folder).glob("*.csv"):
        with Path(file).open(encoding="utf-8") as csv_file:
            if Path(file).name not in ["Withdrawals Report.csv", "Releases Report.csv"]:
                continue

            lines = list(csv.reader(csv_file))
            header = lines[0]
            lines = lines[1:]

            if Path(file).name == "Withdrawals Report.csv":
                _validate_header(header, COLUMNS_WITHDRAWAL, str(file))
                transactions += [
                    _init_from_withdrawal_report(row, str(file)) for row in lines
                ]
            else:
                _validate_header(header, COLUMNS_RELEASE, str(file))
                transactions += [
                    _init_from_release_report(row, str(file)) for row in lines
                ]

    return transactions



#!/usr/bin/env python3
"""Capital Gain Calculator main module."""
from __future__ import annotations

from collections import defaultdict
from dataclasses import astuple
import datetime
import decimal
from decimal import Decimal
import importlib.metadata
import logging
from pathlib import Path
import sys

from . import render_latex
from .args_parser import create_parser
from .const import BED_AND_BREAKFAST_DAYS, CAPITAL_GAIN_ALLOWANCES, INTERNAL_START_DATE
from .currency_converter import CurrencyConverter
from .dates import get_tax_year_end, get_tax_year_start, is_date
from .exceptions import (
    AmountMissingError,
    CalculatedAmountDiscrepancyError,
    CalculationError,
    InvalidTransactionError,
    PriceMissingError,
    QuantityNotPositiveError,
    SymbolMissingError,
)
from .initial_prices import InitialPrices
from .model import (
    ActionType,
    BrokerTransaction,
    CalculationEntry,
    CalculationLog,
    CapitalGainsReport,
    HmrcTransactionLog,
    RuleType,
)
from .parsers import read_broker_transactions, read_initial_prices
from .transaction_log import add_to_list, has_key
from .util import round_decimal

LOGGER = logging.getLogger(__name__)


def get_amount_or_fail(transaction: BrokerTransaction) -> Decimal:
    """Return the transaction amount or throw an error."""
    amount = transaction.amount
    if amount is None:
        raise AmountMissingError(transaction)
    return amount


# It is not clear how Schwab or other brokers round the dollar value,
# so assume the values are equal if they are within $0.01.
def _approx_equal(val_a: Decimal, val_b: Decimal) -> bool:
    return abs(val_a - val_b) < Decimal("0.01")


class CapitalGainsCalculator:
    """Main calculator class."""

    def __init__(
        self,
        tax_year: int,
        converter: CurrencyConverter,
        initial_prices: InitialPrices,
        balance_check: bool = True,
    ):
        """Create calculator object."""
        self.tax_year = tax_year

        self.tax_year_start_date = get_tax_year_start(tax_year)
        self.tax_year_end_date = get_tax_year_end(tax_year)

        self.converter = converter
        self.initial_prices = initial_prices
        self.balance_check = balance_check

    def date_in_tax_year(self, date: datetime.date) -> bool:
        """Check if date is within current tax year."""
        assert is_date(date)
        return self.tax_year_start_date <= date <= self.tax_year_end_date

    def add_acquisition(
        self,
        portfolio: dict[str, Decimal],
        acquisition_list: HmrcTransactionLog,
        transaction: BrokerTransaction,
    ) -> None:
        """Add new acquisition to the given list."""
        symbol = transaction.symbol
        quantity = transaction.quantity
        price = transaction.price
        if symbol is None:
            raise SymbolMissingError(transaction)
        if quantity is None or quantity <= 0:
            raise QuantityNotPositiveError(transaction)
        # This is basically only for data validation
        if symbol in portfolio:
            portfolio[symbol] += quantity
        else:
            portfolio[symbol] = quantity

        # Add to acquisition_list to apply same day rule
        if transaction.action in [ActionType.STOCK_ACTIVITY, ActionType.SPIN_OFF]:
            if price is None:
                price = self.initial_prices.get(transaction.date, symbol)
            amount = round_decimal(quantity * price, 2)
        else:
            if price is None:
                raise PriceMissingError(transaction)

            amount = get_amount_or_fail(transaction)
            calculated_amount = quantity * price + transaction.fees
            if not _approx_equal(amount, -calculated_amount):
                raise CalculatedAmountDiscrepancyError(transaction, -calculated_amount)
            amount = -amount

        add_to_list(
            acquisition_list,
            transaction.date,
            symbol,
            quantity,
            self.converter.to_gbp_for(amount, transaction),
            self.converter.to_gbp_for(transaction.fees, transaction),
        )

    def add_disposal(
        self,
        portfolio: dict[str, Decimal],
        disposal_list: HmrcTransactionLog,
        transaction: BrokerTransaction,
    ) -> None:
        """Add new disposal to the given list."""
        symbol = transaction.symbol
        quantity = transaction.quantity
        if symbol is None:
            raise SymbolMissingError(transaction)
        if symbol not in portfolio:
            raise InvalidTransactionError(
                transaction, "Tried to sell not owned symbol, reversed order?"
            )
        if quantity is None or quantity <= 0:
            raise QuantityNotPositiveError(transaction)
        if portfolio[symbol] < quantity:
            raise InvalidTransactionError(
                transaction,
                f"Tried to sell more than the available balance({portfolio[symbol]})",
            )
        # This is basically only for data validation
        portfolio[symbol] -= quantity
        if portfolio[symbol] == 0:
            del portfolio[symbol]
        # Add to disposal_list to apply same day rule

        amount = get_amount_or_fail(transaction)
        price = transaction.price

        if price is None:
            raise PriceMissingError(transaction)
        calculated_amount = quantity * price - transaction.fees
        if not _approx_equal(amount, calculated_amount):
            raise CalculatedAmountDiscrepancyError(transaction, calculated_amount)
        add_to_list(
            disposal_list,
            transaction.date,
            symbol,
            quantity,
            self.converter.to_gbp_for(amount, transaction),
            self.converter.to_gbp_for(transaction.fees, transaction),
        )

    def convert_to_hmrc_transactions(
        self,
        transactions: list[BrokerTransaction],
    ) -> tuple[HmrcTransactionLog, HmrcTransactionLog]:
        """Convert broker transactions to HMRC transactions."""
        # We keep a balance per broker,currency pair
        balance: dict[tuple[str, str], Decimal] = defaultdict(lambda: Decimal(0))
        dividends = Decimal(0)
        dividends_tax = Decimal(0)
        interest = Decimal(0)
        total_sells = Decimal(0)
        portfolio: dict[str, Decimal] = {}
        acquisition_list: HmrcTransactionLog = {}
        disposal_list: HmrcTransactionLog = {}

        for i, transaction in enumerate(transactions):
            new_balance = balance[(transaction.broker, transaction.currency)]
            if transaction.action is ActionType.TRANSFER:
                new_balance += get_amount_or_fail(transaction)
            elif transaction.action is ActionType.BUY:
                new_balance += get_amount_or_fail(transaction)
                self.add_acquisition(portfolio, acquisition_list, transaction)
            elif transaction.action is ActionType.SELL:
                amount = get_amount_or_fail(transaction)
                new_balance += amount
                self.add_disposal(portfolio, disposal_list, transaction)
                if self.date_in_tax_year(transaction.date):
                    total_sells += self.converter.to_gbp_for(amount, transaction)
            elif transaction.action is ActionType.FEE:
                amount = get_amount_or_fail(transaction)
                new_balance += amount
                transaction.fees = -amount
                transaction.quantity = Decimal(0)
                gbp_fees = self.converter.to_gbp_for(transaction.fees, transaction)
                if transaction.symbol is None:
                    raise SymbolMissingError(transaction)
                add_to_list(
                    acquisition_list,
                    transaction.date,
                    transaction.symbol,
                    transaction.quantity,
                    gbp_fees,
                    gbp_fees,
                )
            elif transaction.action in [
                ActionType.STOCK_ACTIVITY,
                ActionType.SPIN_OFF,
                ActionType.REINVEST_SHARES,
            ]:
                self.add_acquisition(portfolio, acquisition_list, transaction)
            elif transaction.action in [ActionType.DIVIDEND, ActionType.CAPITAL_GAIN]:
                amount = get_amount_or_fail(transaction)
                new_balance += amount
                if self.date_in_tax_year(transaction.date):
                    dividends += self.converter.to_gbp_for(amount, transaction)
            elif transaction.action in [ActionType.TAX, ActionType.ADJUSTMENT]:
                amount = get_amount_or_fail(transaction)
                new_balance += amount
                if self.date_in_tax_year(transaction.date):
                    dividends_tax += self.converter.to_gbp_for(amount, transaction)
            elif transaction.action is ActionType.INTEREST:
                amount = get_amount_or_fail(transaction)
                new_balance += amount
                if self.date_in_tax_year(transaction.date):
                    interest += self.converter.to_gbp_for(amount, transaction)
            elif transaction.action is ActionType.WIRE_FUNDS_RECEIVED:
                amount = get_amount_or_fail(transaction)
                new_balance += amount
            elif transaction.action is ActionType.REINVEST_DIVIDENDS:
                print(f"WARNING: Ignoring unsupported action: {transaction.action}")
            else:
                raise InvalidTransactionError(
                    transaction, f"Action not processed({transaction.action})"
                )
            if self.balance_check and new_balance < 0:
                msg = f"Reached a negative balance({new_balance})"
                msg += f" for broker {transaction.broker} ({transaction.currency})"
                msg += " after processing the following transactions:\n"
                msg += "\n".join(map(str, transactions[: i + 1]))
                raise CalculationError(msg)
            balance[(transaction.broker, transaction.currency)] = new_balance
        print("First pass completed")
        print("Final portfolio:")
        for stock, quantity in portfolio.items():
            print(f"  {stock}: {round_decimal(quantity, 2)}")
        print("Final balance:")
        for (broker, currency), amount in balance.items():
            print(f"  {broker}: {round_decimal(amount, 2)} ({currency})")
        print(f"Dividends: £{round_decimal(dividends, 2)}")
        print(f"Dividend taxes: £{round_decimal(-dividends_tax, 2)}")
        print(f"Interest: £{round_decimal(interest, 2)}")
        print(f"Disposal proceeds: £{round_decimal(total_sells, 2)}")
        print("")
        return acquisition_list, disposal_list

    @staticmethod
    def process_acquisition(
        acquisition_list: HmrcTransactionLog,
        bed_and_breakfast_list: HmrcTransactionLog,
        portfolio: dict[str, tuple[Decimal, Decimal]],
        symbol: str,
        date_index: datetime.date,
    ) -> list[CalculationEntry]:
        """Process single acquisition."""
        acquisition_quantity, acquisition_amount, acquisition_fees = astuple(
            acquisition_list[date_index][symbol]
        )
        original_acquisition_amount = acquisition_amount
        if symbol not in portfolio:
            portfolio[symbol] = (Decimal(0), Decimal(0))
        current_quantity, current_amount = portfolio[symbol]
        calculation_entries = []

        # Management fee transaction can have 0 quantity
        assert acquisition_quantity >= 0
        assert acquisition_amount > 0
        bed_and_breakfast_quantity = Decimal(0)
        bed_and_breakfast_amount = Decimal(0)
        bed_and_breakfast_fees = Decimal(0)
        if acquisition_quantity > 0:
            acquisition_price = acquisition_amount / acquisition_quantity
            if has_key(bed_and_breakfast_list, date_index, symbol):
                (
                    bed_and_breakfast_quantity,
                    bed_and_breakfast_amount,
                    _bed_and_breakfast_fees,
                ) = astuple(bed_and_breakfast_list[date_index][symbol])
                assert bed_and_breakfast_quantity <= acquisition_quantity
                acquisition_amount -= bed_and_breakfast_quantity * acquisition_price
                acquisition_amount += bed_and_breakfast_amount
                assert acquisition_amount > 0
                bed_and_breakfast_fees = (
                    acquisition_fees * bed_and_breakfast_quantity / acquisition_quantity
                )
                calculation_entries.append(
                    CalculationEntry(
                        rule_type=RuleType.BED_AND_BREAKFAST,
                        quantity=bed_and_breakfast_quantity,
                        amount=-bed_and_breakfast_amount,
                        new_quantity=current_quantity + bed_and_breakfast_quantity,
                        new_pool_cost=current_amount + bed_and_breakfast_amount,
                        fees=bed_and_breakfast_fees,
                        allowable_cost=original_acquisition_amount,
                    )
                )
        portfolio[symbol] = (
            current_quantity + acquisition_quantity,
            current_amount + acquisition_amount,
        )
        if (
            acquisition_quantity - bed_and_breakfast_quantity > 0
            or bed_and_breakfast_quantity == 0
        ):
            calculation_entries.append(
                CalculationEntry(
                    rule_type=RuleType.SECTION_104,
                    quantity=acquisition_quantity - bed_and_breakfast_quantity,
                    amount=-(acquisition_amount - bed_and_breakfast_amount),
                    new_quantity=current_quantity + acquisition_quantity,
                    new_pool_cost=current_amount + acquisition_amount,
                    fees=acquisition_fees - bed_and_breakfast_fees,
                    allowable_cost=original_acquisition_amount,
                )
            )
        return calculation_entries

    @staticmethod
    def process_disposal(
        acquisition_list: HmrcTransactionLog,
        disposal_list: HmrcTransactionLog,
        bed_and_breakfast_list: HmrcTransactionLog,
        portfolio: dict[str, tuple[Decimal, Decimal]],
        symbol: str,
        date_index: datetime.date,
    ) -> tuple[Decimal, list[CalculationEntry]]:
        """Process single disposal."""
        disposal_quantity, proceeds_amount, disposal_fees = astuple(
            disposal_list[date_index][symbol]
        )
        original_disposal_quantity = disposal_quantity
        disposal_price = proceeds_amount / disposal_quantity
        current_quantity, current_amount = portfolio[symbol]
        assert disposal_quantity <= current_quantity
        chargeable_gain = Decimal(0)
        calculation_entries = []
        # Same day rule is first
        if has_key(acquisition_list, date_index, symbol):
            same_day_quantity, same_day_amount, _same_day_fees = astuple(
                acquisition_list[date_index][symbol]
            )
            available_quantity = min(disposal_quantity, same_day_quantity)
            if available_quantity > 0:
                acquisition_price = same_day_amount / same_day_quantity
                same_day_proceeds = available_quantity * disposal_price
                same_day_allowable_cost = available_quantity * acquisition_price
                same_day_gain = same_day_proceeds - same_day_allowable_cost
                chargeable_gain += same_day_gain
                LOGGER.debug(
                    "SAME DAY, quantity %d, gain %s, disposal price %s, "
                    "acquisition price %s",
                    available_quantity,
                    same_day_gain,
                    disposal_price,
                    acquisition_price,
                )
                disposal_quantity -= available_quantity
                proceeds_amount -= available_quantity * disposal_price
                current_quantity -= available_quantity
                # These shares shouldn't be added to Section 104 holding
                current_amount -= available_quantity * acquisition_price
                if current_quantity == 0:
                    assert (
                        round_decimal(current_amount, 23) == 0
                    ), f"current amount {current_amount}"
                fees = disposal_fees * available_quantity / original_disposal_quantity
                calculation_entries.append(
                    CalculationEntry(
                        rule_type=RuleType.SAME_DAY,
                        quantity=available_quantity,
                        amount=same_day_proceeds,
                        gain=same_day_gain,
                        allowable_cost=same_day_allowable_cost,
                        fees=fees,
                        new_quantity=current_quantity,
                        new_pool_cost=current_amount,
                    )
                )

        # Bed and breakfast rule next
        if disposal_quantity > 0:
            for i in range(BED_AND_BREAKFAST_DAYS):
                search_index = date_index + datetime.timedelta(days=i + 1)
                if has_key(acquisition_list, search_index, symbol):
                    (
                        acquisition_quantity,
                        acquisition_amount,
                        _acquisition_fees,
                    ) = astuple(acquisition_list[search_index][symbol])

                    bed_and_breakfast_quantity = Decimal(0)
                    if has_key(bed_and_breakfast_list, search_index, symbol):
                        (
                            bed_and_breakfast_quantity,
                            _bb_amount,
                            _bb_fees,
                        ) = astuple(bed_and_breakfast_list[search_index][symbol])
                    assert bed_and_breakfast_quantity <= acquisition_quantity

                    same_day_quantity = Decimal(0)
                    if has_key(disposal_list, search_index, symbol):
                        (
                            same_day_quantity,
                            _same_day_amount,
                            _same_day_fees,
                        ) = astuple(disposal_list[search_index][symbol])
                    assert same_day_quantity <= acquisition_quantity

                    # This can be some management fee entry or already used
                    # by bed and breakfast rule
                    if (
                        acquisition_quantity
                        - same_day_quantity
                        - bed_and_breakfast_quantity
                        == 0
                    ):
                        continue
                    print(
                        f"WARNING: Bed and breakfasting for {symbol}."
                        f" Disposed on {date_index}"
                        f" and acquired again on {search_index}"
                    )
                    available_quantity = min(
                        disposal_quantity,
                        acquisition_quantity
                        - same_day_quantity
                        - bed_and_breakfast_quantity,
                    )
                    acquisition_price = acquisition_amount / acquisition_quantity
                    bed_and_breakfast_proceeds = available_quantity * disposal_price
                    bed_and_breakfast_allowable_cost = (
                        available_quantity * acquisition_price
                    )
                    bed_and_breakfast_gain = (
                        bed_and_breakfast_proceeds - bed_and_breakfast_allowable_cost
                    )
                    chargeable_gain += bed_and_breakfast_gain
                    LOGGER.debug(
                        "BED & BREAKFAST, quantity %d, gain %s, disposal price %s, "
                        "acquisition price %s",
                        available_quantity,
                        bed_and_breakfast_gain,
                        disposal_price,
                        acquisition_price,
                    )
                    disposal_quantity -= available_quantity
                    proceeds_amount -= available_quantity * disposal_price
                    current_price = current_amount / current_quantity
                    amount_delta = available_quantity * current_price
                    current_quantity -= available_quantity
                    current_amount -= amount_delta
                    if current_quantity == 0:
                        assert (
                            round_decimal(current_amount, 23) == 0
                        ), f"current amount {current_amount}"
                    add_to_list(
                        bed_and_breakfast_list,
                        search_index,
                        symbol,
                        available_quantity,
                        amount_delta,
                        Decimal(0),
                    )
                    fees = (
                        disposal_fees * available_quantity / original_disposal_quantity
                    )
                    calculation_entries.append(
                        CalculationEntry(
                            rule_type=RuleType.BED_AND_BREAKFAST,
                            quantity=available_quantity,
                            amount=bed_and_breakfast_proceeds,
                            gain=bed_and_breakfast_gain,
                            allowable_cost=bed_and_breakfast_allowable_cost,
                            fees=fees,
                            bed_and_breakfast_date_index=search_index,
                            new_quantity=current_quantity,
                            new_pool_cost=current_amount,
                        )
                    )
        if disposal_quantity > 0:
            allowable_cost = current_amount * disposal_quantity / current_quantity
            chargeable_gain += proceeds_amount - allowable_cost
            LOGGER.debug(
                "SECTION 104, quantity %d, gain %s, proceeds amount %s, "
                "allowable cost %s",
                disposal_quantity,
                proceeds_amount - allowable_cost,
                proceeds_amount,
                allowable_cost,
            )
            current_quantity -= disposal_quantity
            current_amount -= allowable_cost
            if current_quantity == 0:
                assert (
                    round_decimal(current_amount, 10) == 0
                ), f"current amount {current_amount}"
            fees = disposal_fees * disposal_quantity / original_disposal_quantity
            calculation_entries.append(
                CalculationEntry(
                    rule_type=RuleType.SECTION_104,
                    quantity=disposal_quantity,
                    amount=proceeds_amount,
                    gain=proceeds_amount - allowable_cost,
                    allowable_cost=allowable_cost,
                    fees=fees,
                    new_quantity=current_quantity,
                    new_pool_cost=current_amount,
                )
            )
            disposal_quantity = Decimal(0)

        assert (
            round_decimal(disposal_quantity, 23) == 0
        ), f"disposal quantity {disposal_quantity}"
        portfolio[symbol] = (current_quantity, current_amount)
        chargeable_gain = round_decimal(chargeable_gain, 2)
        return chargeable_gain, calculation_entries

    def calculate_capital_gain(
        self,
        acquisition_list: HmrcTransactionLog,
        disposal_list: HmrcTransactionLog,
    ) -> CapitalGainsReport:
        """Calculate capital gain and return generated report."""
        begin_index = INTERNAL_START_DATE
        tax_year_start_index = self.tax_year_start_date
        end_index = self.tax_year_end_date
        disposal_count = 0
        disposal_proceeds = Decimal(0)
        allowable_costs = Decimal(0)
        capital_gain = Decimal(0)
        capital_loss = Decimal(0)
        bed_and_breakfast_list: HmrcTransactionLog = {}
        portfolio: dict[str, tuple[Decimal, Decimal]] = {}
        calculation_log: CalculationLog = {}
        for date_index in (
            begin_index + datetime.timedelta(days=x)
            for x in range(0, (end_index - begin_index).days + 1)
        ):
            if date_index in acquisition_list:
                for symbol in acquisition_list[date_index]:
                    calculation_entries = self.process_acquisition(
                        acquisition_list,
                        bed_and_breakfast_list,
                        portfolio,
                        symbol,
                        date_index,
                    )
                    if date_index >= tax_year_start_index:
                        if date_index not in calculation_log:
                            calculation_log[date_index] = {}
                        calculation_log[date_index][
                            f"buy${symbol}"
                        ] = calculation_entries
            if date_index in disposal_list:
                for symbol in disposal_list[date_index]:
                    (
                        transaction_capital_gain,
                        calculation_entries,
                    ) = self.process_disposal(
                        acquisition_list,
                        disposal_list,
                        bed_and_breakfast_list,
                        portfolio,
                        symbol,
                        date_index,
                    )
                    if date_index >= tax_year_start_index:
                        disposal_count += 1
                        transaction_disposal_proceeds = disposal_list[date_index][
                            symbol
                        ].amount
                        disposal_proceeds += transaction_disposal_proceeds
                        allowable_costs += (
                            transaction_disposal_proceeds - transaction_capital_gain
                        )
                        transaction_quantity = disposal_list[date_index][
                            symbol
                        ].quantity
                        LOGGER.debug(
                            "DISPOSAL on %s of %s, quantity %d, capital gain $%s",
                            date_index,
                            symbol,
                            transaction_quantity,
                            round_decimal(transaction_capital_gain, 2),
                        )
                        calculated_quantity = Decimal(0)
                        calculated_proceeds = Decimal(0)
                        calculated_gain = Decimal(0)
                        for entry in calculation_entries:
                            calculated_quantity += entry.quantity
                            calculated_proceeds += entry.amount
                            calculated_gain += entry.gain
                        assert transaction_quantity == calculated_quantity
                        assert round_decimal(
                            transaction_disposal_proceeds, 10
                        ) == round_decimal(
                            calculated_proceeds, 10
                        ), f"{transaction_disposal_proceeds} != {calculated_proceeds}"
                        assert transaction_capital_gain == round_decimal(
                            calculated_gain, 2
                        )
                        if date_index not in calculation_log:
                            calculation_log[date_index] = {}
                        calculation_log[date_index][
                            f"sell${symbol}"
                        ] = calculation_entries
                        if transaction_capital_gain > 0:
                            capital_gain += transaction_capital_gain
                        else:
                            capital_loss += transaction_capital_gain
        print("\nSecond pass completed")
        allowance = CAPITAL_GAIN_ALLOWANCES.get(self.tax_year)
        return CapitalGainsReport(
            self.tax_year,
            portfolio,
            disposal_count,
            round_decimal(disposal_proceeds, 2),
            round_decimal(allowable_costs, 2),
            round_decimal(capital_gain, 2),
            round_decimal(capital_loss, 2),
            Decimal(allowance) if allowance is not None else None,
            calculation_log,
        )


def main() -> int:
    """Run main function."""
    # Throw exception on accidental float usage
    decimal.getcontext().traps[decimal.FloatOperation] = True
    args = create_parser().parse_args()

    if args.version:
        print(f"cgt-calc {importlib.metadata.version(__package__)}")
        return 0

    if args.report == "":
        print("error: report name can't be empty")
        return 1

    default_logging_level = logging.DEBUG if args.verbose else logging.WARNING
    logging.basicConfig(level=default_logging_level)

    # Read data from input files
    broker_transactions = read_broker_transactions(
        args.schwab, args.schwab_award, args.trading212, args.mssb, args.sharesight
    )
    converter = CurrencyConverter(args.exchange_rates_file)
    initial_prices = InitialPrices(read_initial_prices(args.initial_prices))

    calculator = CapitalGainsCalculator(
        args.year, converter, initial_prices, balance_check=args.balance_check
    )
    # First pass converts broker transactions to HMRC transactions.
    # This means applying same day rule and collapsing all transactions with
    # same type within the same day.
    # It also converts prices to GBP, validates data and calculates dividends,
    # taxes on dividends and interest.
    acquisition_list, disposal_list = calculator.convert_to_hmrc_transactions(
        broker_transactions
    )
    # Second pass calculates capital gain tax for the given tax year.
    report = calculator.calculate_capital_gain(acquisition_list, disposal_list)
    print(report)

    # Generate PDF report.
    if not args.no_report:
        render_latex.render_calculations(
            report,
            output_path=Path(args.report),
            skip_pdflatex=args.no_pdflatex,
        )
    print("All done!")

    return 0


def init() -> None:
    """Entry point."""
    sys.exit(main())


if __name__ == "__main__":
    init()



"""Model classes."""
from __future__ import annotations

from dataclasses import dataclass
import datetime
from decimal import Decimal
from enum import Enum
from typing import Dict, List

from .util import round_decimal


@dataclass
class HmrcTransactionData:
    """Hmrc transaction figures."""

    quantity: Decimal
    amount: Decimal
    fees: Decimal


# For mapping of dates to int
HmrcTransactionLog = Dict[datetime.date, Dict[str, HmrcTransactionData]]


class ActionType(Enum):
    """Type of transaction action."""

    BUY = 1
    SELL = 2
    TRANSFER = 3
    STOCK_ACTIVITY = 4
    DIVIDEND = 5
    TAX = 6
    FEE = 7
    ADJUSTMENT = 8
    CAPITAL_GAIN = 9
    SPIN_OFF = 10
    INTEREST = 11
    REINVEST_SHARES = 12
    REINVEST_DIVIDENDS = 13
    WIRE_FUNDS_RECEIVED = 14


@dataclass
class BrokerTransaction:
    """Broken transaction data."""

    date: datetime.date
    action: ActionType
    symbol: str | None
    description: str
    quantity: Decimal | None
    price: Decimal | None
    fees: Decimal
    amount: Decimal | None
    currency: str
    broker: str


class RuleType(Enum):
    """HMRC rule type."""

    SECTION_104 = 1
    SAME_DAY = 2
    BED_AND_BREAKFAST = 3


class CalculationEntry:  # noqa: SIM119 # this has non-trivial constructor
    """Calculation entry for final report."""

    def __init__(
        self,
        rule_type: RuleType,
        quantity: Decimal,
        amount: Decimal,
        fees: Decimal,
        new_quantity: Decimal,
        new_pool_cost: Decimal,
        gain: Decimal | None = None,
        allowable_cost: Decimal | None = None,
        bed_and_breakfast_date_index: datetime.date | None = None,
    ):
        """Create calculation entry."""
        self.rule_type = rule_type
        self.quantity = quantity
        self.amount = amount
        self.allowable_cost = (
            allowable_cost if allowable_cost is not None else Decimal(0)
        )
        self.fees = fees
        self.gain = gain if gain is not None else Decimal(0)
        self.new_quantity = new_quantity
        self.new_pool_cost = new_pool_cost
        self.bed_and_breakfast_date_index = bed_and_breakfast_date_index
        if self.amount >= 0:
            assert self.gain == self.amount - self.allowable_cost

    def __repr__(self) -> str:
        """Return print representation."""
        return f"<CalculationEntry {str(self)}>"

    def __str__(self) -> str:
        """Return string representation."""
        return (
            f"{self.rule_type.name.replace('_', ' ')}, "
            f"quantity: {self.quantity}, "
            f"disposal proceeds: {self.amount}, "
            f"allowable cost: {self.allowable_cost}, "
            f"fees: {self.fees}, "
            f"gain: {self.gain}"
        )


CalculationLog = Dict[datetime.date, Dict[str, List[CalculationEntry]]]


@dataclass
class CapitalGainsReport:
    """Store calculated report."""

    tax_year: int
    portfolio: dict[str, tuple[Decimal, Decimal]]
    disposal_count: int
    disposal_proceeds: Decimal
    allowable_costs: Decimal
    capital_gain: Decimal
    capital_loss: Decimal
    capital_gain_allowance: Decimal | None
    calculation_log: CalculationLog

    def total_gain(self) -> Decimal:
        """Total capital gain."""
        return self.capital_gain + self.capital_loss

    def taxable_gain(self) -> Decimal:
        """Taxable gain with current allowance."""
        assert self.capital_gain_allowance is not None
        return max(Decimal(0), self.total_gain() - self.capital_gain_allowance)

    def __repr__(self) -> str:
        """Return string representation."""
        return f"<CalculationEntry: {str(self)}>"

    def __str__(self) -> str:
        """Return string representation."""
        out = f"Portfolio at the end of {self.tax_year}/{self.tax_year + 1} tax year:\n"
        for symbol, (quantity, amount) in self.portfolio.items():
            if quantity > 0:
                out += (
                    f"  {symbol}: {round_decimal(quantity, 2)}, "
                    f"£{round_decimal(amount, 2)}\n"
                )
        out += f"For tax year {self.tax_year}/{self.tax_year + 1}:\n"
        out += f"Number of disposals: {self.disposal_count}\n"
        out += f"Disposal proceeds: £{self.disposal_proceeds}\n"
        out += f"Allowable costs: £{self.allowable_costs}\n"
        out += f"Capital gain: £{self.capital_gain}\n"
        out += f"Capital loss: £{-self.capital_loss}\n"
        out += f"Total capital gain: £{self.total_gain()}\n"
        if self.capital_gain_allowance is not None:
            out += f"Taxable capital gain: £{self.taxable_gain()}\n"
        else:
            out += "WARNING: Missing allowance for this tax year\n"
        return out



"""Parse command line arguments."""
import argparse
import datetime

from .const import DEFAULT_EXCHANGE_RATES_FILE, DEFAULT_REPORT_PATH


def get_last_elapsed_tax_year() -> int:
    """Get last ended tax year."""
    now = datetime.datetime.now()
    if now.date() >= datetime.date(now.year, 4, 6):
        return now.year - 1
    return now.year - 2


def create_parser() -> argparse.ArgumentParser:
    """Create ArgumentParser."""
    parser = argparse.ArgumentParser(
        description="Calculate capital gains from stock transactions.",
    )
    parser.add_argument(
        "--year",
        type=int,
        default=get_last_elapsed_tax_year(),
        nargs="?",
        help="First year of the tax year to calculate gains on (default: %(default)d)",
    )
    parser.add_argument(
        "--schwab",
        type=str,
        nargs="?",
        help="file containing the exported transactions from Charles Schwab",
    )
    parser.add_argument(
        "--schwab-award",
        type=str,
        default=None,
        nargs="?",
        help="file containing schwab award data for stock prices",
    )
    parser.add_argument(
        "--trading212",
        type=str,
        nargs="?",
        help="folder containing the exported transaction files from Trading 212",
    )
    parser.add_argument(
        "--mssb",
        type=str,
        nargs="?",
        help="folder containing the exported transaction files from Morgan Stanley",
    )
    parser.add_argument(
        "--sharesight",
        type=str,
        nargs="?",
        help="folder containing reports from Sharesight in CSV format",
    )

    parser.add_argument(
        "--exchange-rates-file",
        type=str,
        default=DEFAULT_EXCHANGE_RATES_FILE,
        nargs="?",
        help="output file for monthly exchange rates from HMRC",
    )
    parser.add_argument(
        "--initial-prices",
        type=str,
        default=None,
        nargs="?",
        help="file containing stock prices in USD at the moment of vesting, split, etc",
    )
    parser.add_argument(
        "--report",
        type=str,
        default=DEFAULT_REPORT_PATH,
        nargs="?",
        help="where to save the generated PDF report (default: %(default)s)",
    )
    parser.add_argument(
        "--no-report",
        action="store_true",
        help="do not generate PDF report",
    )
    parser.add_argument(
        "--no-balance-check",
        dest="balance_check",
        action="store_false",
        default=True,
    )
    parser.add_argument(
        "--verbose",
        action="store_true",
        help="enable extra logging",
    )
    parser.add_argument(
        "--version",
        action="store_true",
        help="print version",
    )
    # For testing only
    parser.add_argument(
        "--no-pdflatex",
        action="store_true",
        help=argparse.SUPPRESS,
    )
    return parser



"""Sharesight parser."""
from __future__ import annotations

import csv
from datetime import date, datetime
from decimal import Decimal, InvalidOperation
from pathlib import Path
from typing import Final, Iterable, Iterator, List

from cgt_calc.exceptions import InvalidTransactionError, ParsingError
from cgt_calc.model import ActionType, BrokerTransaction

STOCK_ACTIVITY_COMMENT_MARKER: Final[str] = "Stock Activity"


def parse_date(val: str) -> date:
    """Parse a Sharesight report date."""

    return datetime.strptime(val, "%d/%m/%Y").date()


def parse_decimal(val: str) -> Decimal:
    """Convert value to Decimal."""
    try:
        return Decimal(val.replace(",", ""))
    except InvalidOperation:
        raise ValueError(f"Bad decimal: {val}") from None


def maybe_decimal(val: str) -> Decimal | None:
    """Convert value to Decimal."""
    return parse_decimal(val) if val else None


class SharesightTransaction(BrokerTransaction):
    """Sharesight transaction.

    Just a marker type for now
    """


class RowIterator(Iterator[List[str]]):
    """Iterator for CSV rows that keeps track of line number."""

    def __init__(self, rows: Iterable[list[str]]) -> None:
        """Initialise RowIterator."""
        self.rows = iter(rows)
        self.line = 1

    def __next__(self) -> list[str]:
        """Produce next element and increment line number."""
        elm = next(self.rows)
        self.line += 1
        return elm

    def __iter__(self) -> RowIterator:
        """Return an iterator for this object."""
        return self


def parse_dividend_payments(
    rows: Iterator[list[str]],
) -> Iterable[SharesightTransaction]:
    """Parse dividend payments from Sharesight data.

    This is the section started with a "Foreign Income" or "Local Income" header.
    We parse those two sections very similarly, so we use one function.
    """

    columns = next(rows, None)
    if columns is None:
        return

    for row in rows:
        if row[0] == "Total":
            # Don't use the totals row, but it signals the end of the section
            break

        row_dict = dict(zip(columns, row))

        dividend_date = parse_date(row_dict["Date Paid"])
        symbol = row_dict["Code"]
        description = row_dict["Comments"]
        broker = "Sharesight"

        currency = row_dict.get("Currency")
        # If we have a currency this is foreign income, otherwise it's local
        if currency:
            amount = parse_decimal(row_dict["Gross Amount"])
            tax = maybe_decimal(row_dict["Foreign Tax Deducted"])
        else:
            amount = parse_decimal(row_dict["Gross Dividend"])
            tax = maybe_decimal(row_dict["Tax Deducted"])
            # Local income must be in GBP, otherwise why are you using this tool?
            currency = "GBP"

        yield SharesightTransaction(
            date=dividend_date,
            action=ActionType.DIVIDEND,
            symbol=symbol,
            description=description,
            broker=broker,
            currency=currency,
            amount=amount,
            quantity=None,
            price=None,
            fees=Decimal(0),
        )

        # Generate the tax as a separate transaction
        if tax:
            yield SharesightTransaction(
                date=dividend_date,
                action=ActionType.TAX,
                symbol=symbol,
                description=description,
                broker=broker,
                currency=currency,
                amount=-tax,
                quantity=None,
                price=None,
                fees=Decimal(0),
            )


def parse_local_income(rows: Iterator[list[str]]) -> Iterable[SharesightTransaction]:
    """Parse Local Income section from Sharesight data.

    This basically just yields to `parse_dividend_payments`, but we skip the
    header lines
    """

    for row in rows:
        if row[0] == "Total Local Income":
            return

        if row[0] == "Dividend Payments":
            yield from parse_dividend_payments(rows)


def parse_foreign_income(rows: Iterator[list[str]]) -> Iterable[SharesightTransaction]:
    """Parse Foreign Income section from Sharesight data."""

    yield from parse_dividend_payments(rows)


def parse_income_report(file: Path) -> Iterable[SharesightTransaction]:
    """Parse the Taxable Income Report from Sharesight."""

    with file.open(encoding="utf-8") as csv_file:
        rows = list(csv.reader(csv_file))

    # Use our custom iterator for error reporting
    rows_iter = RowIterator(rows)
    for row in rows_iter:
        try:
            if row[0] == "Local Income":
                yield from parse_local_income(rows_iter)
            elif row[0] == "Foreign Income":
                yield from parse_foreign_income(rows_iter)
        except ValueError as err:
            raise ParsingError(f"{file}:{rows_iter.line}", str(err)) from None


def parse_trades(
    columns: list[str], rows: Iterator[list[str]]
) -> Iterable[SharesightTransaction]:
    """Parse content in All Trades Report from Sharesight."""

    for row in rows:
        if not any(row):
            # There is an empty row at the end of the trades list
            break

        row_dict = dict(zip(columns, row))
        tpe = row_dict["Type"]
        if tpe == "Buy":
            action = ActionType.BUY
        elif tpe == "Sell":
            action = ActionType.SELL
        else:
            raise ValueError(f"Unknown action: {tpe}")

        market = row_dict["Market"]
        symbol = f"{market}:{row_dict['Code']}"
        trade_date = parse_date(row_dict["Date"])
        quantity = parse_decimal(row_dict["Quantity"])
        price = parse_decimal(row_dict["Price *"])
        fees = maybe_decimal(row_dict["Brokerage *"]) or Decimal(0)
        currency = row_dict["Currency"]
        description = row_dict["Comments"]
        broker = "Sharesight"
        gbp_value = maybe_decimal(row_dict["Value"])

        # Sharesight's reports conventions are slightly different from our
        # conventions:
        # - Quantity is negative on sell
        # - Amount is negative when selling, and positive when buying
        #   (as it tracks portfolio value, not account balance)
        # - Value provided is always in GBP
        # - Foreign exchange transactions are show as BUY and SELL

        if market == "FX":
            # While Sharesight provides an exchange rate, it is not precise enough
            # for cryptocurrency transactions
            if not gbp_value:
                raise ValueError("Missing Value in FX transaction")

            price = abs(gbp_value / quantity)
            currency = "GBP"

        # Make amount positive on sell and negative on buy
        amount = -(quantity * price) - fees
        # Make quantity always positive
        quantity = abs(quantity)

        # Create the transaction object now so we can report it in exceptions
        transaction = SharesightTransaction(
            date=trade_date,
            action=action,
            symbol=symbol,
            description=description,
            quantity=quantity,
            price=price,
            fees=fees,
            amount=amount,
            currency=currency,
            broker=broker,
        )

        # Sharesight has no native support for stock activity, so use a string
        # in the trade comment to mark it
        if STOCK_ACTIVITY_COMMENT_MARKER.lower() in description.lower():
            # Stock activity that is not a grant is weird and unsupported
            if action != ActionType.BUY:
                raise InvalidTransactionError(
                    transaction, "Stock activity must have Type=Buy"
                )

            transaction.action = ActionType.STOCK_ACTIVITY
            transaction.amount = None

        yield transaction


def parse_trade_report(file: Path) -> Iterable[SharesightTransaction]:
    """Parse All Trades Report from Sharesight."""

    with file.open(encoding="utf-8") as csv_file:
        rows = list(csv.reader(csv_file))

    # Use our custom iterator for error reporting
    rows_iter = RowIterator(rows)
    for row in rows_iter:
        # Skip everything until we find the header
        if row[0] == "Market":
            columns = row
            try:
                yield from parse_trades(columns, rows_iter)
            except (InvalidOperation, ValueError) as err:
                raise ParsingError(f"{file}:{rows_iter.line}", str(err)) from None


def read_sharesight_transactions(
    transactions_folder: str,
) -> list[SharesightTransaction]:
    """Parse Sharesight transactions from reports."""

    transactions: list[SharesightTransaction] = []
    for file in Path(transactions_folder).glob("*.csv"):
        if file.match("Taxable Income Report*.csv"):
            income_transactions = list(parse_income_report(file))
            if not income_transactions:
                print(f"WARNING: no transactions detected in file {file}")
            else:
                transactions += income_transactions

        if file.match("All Trades Report*.csv"):
            trade_transactions = list(parse_trade_report(file))
            if not trade_transactions:
                print(f"WARNING: no transactions detected in file {file}")
            else:
                transactions += trade_transactions

    def key(transaction: SharesightTransaction) -> date:
        return transaction.date

    transactions.sort(key=key)
    return transactions



"""Initial stock prices."""
from __future__ import annotations

from dataclasses import dataclass
import datetime
from decimal import Decimal

from .dates import is_date
from .exceptions import ExchangeRateMissingError


@dataclass
class InitialPrices:
    """Class to store initial stock prices."""

    initial_prices: dict[datetime.date, dict[str, Decimal]]

    def get(self, date: datetime.date, symbol: str) -> Decimal:
        """Get initial stock price at given date."""
        assert is_date(date)
        if date not in self.initial_prices or symbol not in self.initial_prices[date]:
            raise ExchangeRateMissingError(symbol, date)
        return self.initial_prices[date][symbol]



"""Constants."""
from __future__ import annotations

import datetime
from typing import Final

# Allowances from
# https://www.gov.uk/guidance/capital-gains-tax-rates-and-allowances#tax-free-allowances-for-capital-gains-tax
CAPITAL_GAIN_ALLOWANCES: Final[dict[int, int]] = {
    2014: 11000,
    2015: 11100,
    2016: 11100,
    2017: 11300,
    2018: 11700,
    2019: 12000,
    2020: 12300,
    2021: 12300,
}

DEFAULT_REPORT_PATH: Final = "calculations.pdf"

INTERNAL_START_DATE: Final = datetime.date(2010, 1, 1)

# Resources

PACKAGE_NAME = __package__

# Monthly exchange rate history from HMRC
DEFAULT_EXCHANGE_RATES_FILE: Final = "exchange_rates.csv"

# Initial vesting and spin-off prices
DEFAULT_INITIAL_PRICES_FILE: Final = "initial_prices.csv"

# Latex template for calculations report
TEMPLATE_NAME: Final = "template.tex.j2"

BED_AND_BREAKFAST_DAYS: Final = 30



"""Function to work with dates."""
import datetime


def is_date(date: datetime.date) -> bool:
    """Check if date has only date but not time."""
    if not isinstance(date, datetime.date) or isinstance(date, datetime.datetime):
        raise Exception(f'should be datetime.date: {type(date)} "{date}"')
    return True


def get_tax_year_start(tax_year: int) -> datetime.date:
    """Return tax year start date."""
    # 6 April
    return datetime.date(tax_year, 4, 6)


def get_tax_year_end(tax_year: int) -> datetime.date:
    """Return tax year end date."""
    # 5 April
    return datetime.date(tax_year + 1, 4, 5)



"""Trading 212 parser."""
from __future__ import annotations

import csv
from datetime import datetime
from decimal import Decimal
from pathlib import Path
from typing import Final

from cgt_calc.exceptions import ParsingError
from cgt_calc.model import ActionType, BrokerTransaction

COLUMNS: Final[list[str]] = [
    "Action",
    "Time",
    "ISIN",
    "Ticker",
    "Name",
    "No. of shares",
    "Price / share",
    "Currency (Price / share)",
    "Exchange rate",
    "Result (GBP)",
    "Total (GBP)",
    "Withholding tax",
    "Currency (Withholding tax)",
    "Charge amount (GBP)",
    "Transaction fee (GBP)",
    "Finra fee (GBP)",
    "Notes",
    "ID",
    "Currency conversion fee (GBP)",
]


def decimal_or_none(val: str) -> Decimal | None:
    """Convert value to Decimal."""
    return Decimal(val) if val not in ["", "Not available"] else None


def action_from_str(label: str, filename: str) -> ActionType:
    """Convert label to ActionType."""
    if label in [
        "Market buy",
        "Limit buy",
    ]:
        return ActionType.BUY

    if label in [
        "Market sell",
        "Limit sell",
    ]:
        return ActionType.SELL

    if label in [
        "Deposit",
        "Withdrawal",
    ]:
        return ActionType.TRANSFER

    if label in ["Dividend (Ordinary)"]:
        return ActionType.DIVIDEND

    raise ParsingError(filename, f"Unknown action: {label}")


class Trading212Transaction(BrokerTransaction):
    """Represent single Trading 212 transaction."""

    def __init__(self, header: list[str], row_raw: list[str], filename: str):
        """Create transaction from CSV row."""
        row = dict(zip(header, row_raw))
        time_str = row["Time"]
        self.datetime = datetime.strptime(time_str, "%Y-%m-%d %H:%M:%S")
        date = self.datetime.date()
        self.raw_action = row["Action"]
        action = action_from_str(self.raw_action, filename)
        symbol = row["Ticker"] if row["Ticker"] != "" else None
        description = row["Name"]
        quantity = decimal_or_none(row["No. of shares"])
        self.price_foreign = decimal_or_none(row["Price / share"])
        self.currency_foreign = row["Currency (Price / share)"]
        self.exchange_rate = decimal_or_none(row["Exchange rate"])
        self.transaction_fee = decimal_or_none(row.get("Transaction fee (GBP)", "0"))
        self.finra_fee = decimal_or_none(row.get("Finra fee (GBP)", "0"))
        self.conversion_fee = decimal_or_none(
            row.get("Currency conversion fee (GBP)", "0")
        )
        fees = (
            (self.transaction_fee or Decimal(0))
            + (self.finra_fee or Decimal(0))
            + (self.conversion_fee or Decimal(0))
        )
        amount = decimal_or_none(row["Total (GBP)"])
        price = (
            abs(amount / quantity)
            if amount is not None and quantity is not None
            else None
        )
        if amount is not None:
            if action == ActionType.BUY or self.raw_action == "Withdrawal":
                amount *= -1
            amount -= fees
        self.isin = row["ISIN"]
        self.transaction_id = row["ID"]
        self.notes = row["Notes"]
        broker = "Trading212"
        super().__init__(
            date,
            action,
            symbol,
            description,
            quantity,
            price,
            fees,
            amount,
            "GBP",
            broker,
        )

    def __eq__(self, other: object) -> bool:
        """Compare transactions by ID."""
        if not isinstance(other, Trading212Transaction):
            raise NotImplementedError()
        return self.transaction_id == other.transaction_id

    def __hash__(self) -> int:
        """Calculate hash."""
        return hash(self.transaction_id)


def validate_header(header: list[str], filename: str) -> None:
    """Check if header is valid."""
    for actual in header:
        if actual not in COLUMNS:
            msg = f"Unknown column {actual}"
            raise ParsingError(filename, msg)


def by_date_and_action(transaction: Trading212Transaction) -> tuple[datetime, bool]:
    """Sort by date and action type."""

    # If there's a deposit in the same second as a buy
    # (happens with the referral award at least)
    # we want to put the buy last to avoid negative balance errors
    return (transaction.datetime, transaction.action == ActionType.BUY)


def read_trading212_transactions(transactions_folder: str) -> list[BrokerTransaction]:
    """Parse Trading 212 transactions from CSV file."""
    transactions = []
    for file in Path(transactions_folder).glob("*.csv"):
        with Path(file).open(encoding="utf-8") as csv_file:
            print(f"Parsing {file}")
            lines = list(csv.reader(csv_file))
            header = lines[0]
            validate_header(header, str(file))
            lines = lines[1:]
            cur_transactions = [
                Trading212Transaction(header, row, str(file)) for row in lines
            ]
            if len(cur_transactions) == 0:
                print(f"WARNING: no transactions detected in file {file}")
            transactions += cur_transactions
    # remove duplicates
    transactions = list(set(transactions))
    transactions.sort(key=by_date_and_action)
    return list(transactions)



"""Convert currencies to GBP using rate history."""
from __future__ import annotations

import csv
import datetime
from decimal import Decimal
from pathlib import Path
from typing import Final
from xml.etree import ElementTree

import requests

from .dates import is_date
from .exceptions import ExchangeRateMissingError, ParsingError
from .model import BrokerTransaction

EXCHANGE_RATES_HEADER: Final = ["month", "currency", "rate"]


class CurrencyConverter:
    """Coverter which holds rate history."""

    def __init__(
        self,
        exchange_rates_file: str | None = None,
        initial_data: dict[str, dict[str, Decimal]] | None = None,
    ):
        """Load data from exchange_rates_file and optionally from initial_data."""
        self.exchange_rates_file = exchange_rates_file
        read_data = self._read_exchange_rates_file(exchange_rates_file)
        self.cache = {**read_data, **(initial_data or {})}

    @staticmethod
    def _read_exchange_rates_file(
        exchange_rates_file: str | None,
    ) -> dict[str, dict[str, Decimal]]:
        cache: dict[str, dict[str, Decimal]] = {}
        if exchange_rates_file is None:
            return cache
        path = Path(exchange_rates_file)
        if not path.is_file():
            return cache
        with path.open(encoding="utf8") as fin:
            csv_reader = csv.DictReader(fin)
            # skip the header
            next(csv_reader)
            for line in csv_reader:
                if sorted(EXCHANGE_RATES_HEADER) != sorted(line.keys()):
                    raise ParsingError(
                        exchange_rates_file,
                        f"invalid columns {line.keys()},"
                        f"they should be {EXCHANGE_RATES_HEADER}",
                    )
                if line["month"] not in cache:
                    cache[line["month"]] = {}
                cache[line["month"]][line["currency"]] = Decimal(line["rate"])
            return cache

    @staticmethod
    def _write_exchange_rates_file(
        exchange_rates_file: str | None, data: dict[str, dict[str, Decimal]]
    ) -> None:
        if exchange_rates_file is None:
            return
        with Path(exchange_rates_file).open("w", encoding="utf8") as fout:
            data_rows = [
                [month, symbol, str(rate)]
                for month, rates in data.items()
                for symbol, rate in rates.items()
            ]
            writer = csv.writer(fout)
            writer.writerows([EXCHANGE_RATES_HEADER] + data_rows)

    def _query_hmrc_api(self, month_str: str) -> None:
        url = (
            "http://www.hmrc.gov.uk/softwaredevelopers/rates/"
            f"exrates-monthly-{month_str}.xml"
        )
        response = requests.get(url)
        if not response.ok:
            raise ParsingError(
                url, f"HMRC API returned a {response.status_code} response"
            )
        tree = ElementTree.fromstring(response.text)
        rates = {
            str(getattr(row.find("currencyCode"), "text", None)).upper(): Decimal(
                str(getattr(row.find("rateNew"), "text", None))
            )
            for row in tree
        }
        if None in rates or None in rates.values():
            raise ParsingError(url, "HMRC API produced invalid/unknown data")
        self.cache[month_str] = rates
        self._write_exchange_rates_file(self.exchange_rates_file, self.cache)

    def currency_to_gbp_rate(self, currency: str, date: datetime.date) -> Decimal:
        """Get GBP/currency rate at given date."""
        assert is_date(date)
        month_str = date.strftime("%m%y")
        if month_str not in self.cache:
            self._query_hmrc_api(month_str)
        if currency not in self.cache[month_str]:
            raise ExchangeRateMissingError(currency, date)
        return self.cache[month_str][currency]

    def to_gbp(self, amount: Decimal, currency: str, date: datetime.date) -> Decimal:
        """Convert amount from given currency to GBP."""
        if currency == "GBP":
            return amount
        return amount / self.currency_to_gbp_rate(currency.upper(), date)

    def to_gbp_for(self, amount: Decimal, transaction: BrokerTransaction) -> Decimal:
        """Convert amount from transaction currency to GBP."""

        return self.to_gbp(amount, transaction.currency, transaction.date)



"""Utility functions."""
import decimal
from decimal import Decimal


def round_decimal(value: Decimal, digits: int = 0) -> Decimal:
    """Round decimal to given precision."""
    with decimal.localcontext() as ctx:
        ctx.rounding = decimal.ROUND_HALF_UP
        return Decimal(round(value, digits))


def strip_zeros(value: Decimal) -> str:
    """Strip trailing zeros from Decimal."""
    return f"{value:.10f}".rstrip("0").rstrip(".")



"""Exceptions to different errors."""
from __future__ import annotations

import datetime
from decimal import Decimal

from .model import BrokerTransaction


class ParsingError(Exception):
    """Parsing error."""

    def __init__(self, file: str, message: str):
        """Initialise."""
        self.message = f"While parsing {file}, {message}"
        super().__init__(self.message)


class InvalidTransactionError(Exception):
    """Invalid transaction error."""

    def __init__(self, transaction: BrokerTransaction, message: str):
        """Initialise."""
        self.transaction = transaction
        self.message = f"{message} for the following transaction:\n{transaction}"
        super().__init__(self.message)


class AmountMissingError(InvalidTransactionError):
    """Amount is missing error."""

    def __init__(self, transaction: BrokerTransaction):
        """Initialise."""
        super().__init__(transaction, "Amount missing")


class SymbolMissingError(InvalidTransactionError):
    """Symbol is missing error."""

    def __init__(self, transaction: BrokerTransaction):
        """Initialise."""
        super().__init__(transaction, "Symbol missing")


class PriceMissingError(InvalidTransactionError):
    """Price is missing error."""

    def __init__(self, transaction: BrokerTransaction):
        """Initialise."""
        super().__init__(transaction, "Price missing")


class QuantityNotPositiveError(InvalidTransactionError):
    """Quantity is negative error."""

    def __init__(self, transaction: BrokerTransaction):
        """Initialise."""
        super().__init__(transaction, "Positive quantity required")


class UnexpectedColumnCountError(ParsingError):
    """Unexpected column error."""

    def __init__(self, row: list[str], count: int, file: str):
        """Initialise."""
        super().__init__(
            file, f"The following row doesn't have {count} columns:\n{row}"
        )


class UnexpectedRowCountError(ParsingError):
    """Unexpected row error."""

    def __init__(self, count: int, file: str):
        """Initialise."""
        super().__init__(file, f"The following file doesn't have {count} rows:")


class CalculatedAmountDiscrepancyError(InvalidTransactionError):
    """Calculated amount discrepancy error."""

    def __init__(self, transaction: BrokerTransaction, calculated_amount: Decimal):
        """Initialise."""
        super().__init__(
            transaction,
            (
                f"Calculated amount({calculated_amount}) differs "
                f"from supplied amount ({transaction.amount})"
            ),
        )


class CalculationError(Exception):
    """Calculation error."""


class ExchangeRateMissingError(CalculationError):
    """Exchange rate is missing error."""

    def __init__(self, symbol: str, date: datetime.date):
        """Initialise."""
        self.message = f"No GBP/{symbol} price for {date}"
        super().__init__(self.message)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cgt_calc', 'cgt_calc.parsers', 'cgt_calc.resources']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=2.11.3,<4.0.0',
 'requests>=2.27.1,<3.0.0',
 'types-requests>=2.27.7,<3.0.0']

entry_points = \
{'console_scripts': ['cgt-calc = cgt_calc.main:init']}

setup_kwargs = {
    'name': 'cgt-calc',
    'version': '1.6.0',
    'description': 'UK capital gains tax calculator for Charles Schwab and Trading 212 accounts',
    'long_description': '# UK capital gains calculator\n\n[![CI](https://github.com/KapJI/capital-gains-calculator/actions/workflows/ci.yml/badge.svg)](https://github.com/KapJI/capital-gains-calculator/actions)\n[![PyPI version](https://img.shields.io/pypi/v/cgt-calc)](https://pypi.org/project/cgt-calc/)\n\nCalculate capital gains tax by transaction history exported from Charles Schwab, Trading 212 and Morgan Stanley. Generate PDF report with calculations.\n\nAutomatically convert all prices to GBP and apply HMRC rules to calculate capital gains tax: "same day" rule, "bed and breakfast" rule, section 104 holding.\n\n## Report example\n\n[calculations_example.pdf](https://github.com/KapJI/capital-gains-calculator/blob/main/calculations_example.pdf)\n\n## Installation\n\nInstall it with [pipx](https://pypa.github.io/pipx/) (or regular pip):\n\n```shell\npipx install cgt-calc\n```\n\n## Prerequisites\n\n-   Python 3.8 or above.\n-   `pdflatex` is required to generate the report.\n\n## Install LaTeX\n\n### MacOS\n\n```shell\nbrew install --cask mactex-no-gui\n```\n\n### Debian based\n\n```shell\napt install texlive-latex-base\n```\n\n### Windows\n\n[Install MiKTeX.](https://miktex.org/download)\n\n## Usage\n\nYou will need several input files:\n\n-   Exported transaction history from Schwab in CSV format since the beginning.\n    Or at least since you first acquired the shares, which you were holding during the tax year.\n    [See example](https://github.com/KapJI/capital-gains-calculator/blob/main/tests/test_data/schwab_transactions.csv).\n-   Exported transaction history from Trading 212.\n    You can use several files here since Trading 212 limit the statements to 1 year periods.\n    [See example](https://github.com/KapJI/capital-gains-calculator/tree/main/tests/test_data/trading212).\n-   Exported transaction history from Morgan Stanley.\n    Since Morgan Stanley generates multiple files in a single report, please specify a directory produced from the report download page.\n-   Exported transaction history from Sharesight\n    Sharesight is a portfolio tracking tool with support for multiple brokers.\n\n    You will need the "All Trades" and "Taxable Income" reports since the beginning.\n    Make sure to select "Since Inception" for the period, and "Not Grouping".\n    Export both reports to Excel or Google Sheets, save as CSV, and place them in the same folder.\n\n    Sharesight aggregates transactions from multiple brokers, but doesn\'t necessarily have balance information.\n    Use the `--no-balance-check` flag to avoid spurious errors.\n\n    Since there is no direct support for equity grants, add `Stock Activity` as part of the comment associated with any vesting transactions - making sure they have the grant price filled.\n\n    [See example](https://github.com/KapJI/capital-gains-calculator/tree/main/tests/test_data/sharesight).\n\n-   CSV file with initial stock prices in USD at the moment of vesting, split, etc.\n    [`initial_prices.csv`](https://github.com/KapJI/capital-gains-calculator/blob/main/cgt_calc/resources/initial_prices.csv) comes pre-packaged, you need to use the same format.\n-   (Optional) Monthly exchange rates prices from [gov.uk](https://www.gov.uk/government/collections/exchange-rates-for-customs-and-vat).\n    `exchange_rates.csv` gets generated automatically using HMRC API, you need to use the same format if you want to override it.\n\nThen run (you can omit the brokers you don\'t use):\n\n```shell\ncgt-calc --year 2020 --schwab schwab_transactions.csv --trading212 trading212/ --mssb mmsb_report/\n```\n\nSee `cgt-calc --help` for the full list of settings.\n\n## Disclaimer\n\nPlease be aware that I\'m not a tax adviser so use this data at your own risk.\n\n## Contribute\n\nAll contributions are highly welcomed.\nIf you notice any bugs please open an issue or send a PR to fix it.\n\nFeel free to add new parsers to support transaction history from more brokers.\n\n## Testing\n\nThis project uses [Poetry](https://python-poetry.org/) for managing dependencies.\n\n-   For local testing you need to [install it](https://python-poetry.org/docs/#installation).\n-   After that run `poetry install` to install all dependencies.\n-   Then activate `pre-commit` hook: `poetry run pre-commit install`\n\nYou can also run all linters and tests manually with this command:\n\n```shell\npoetry run pre-commit run --all-files\n```\n',
    'author': 'Ruslan Sayfutdinov',
    'author_email': 'ruslan@sayfutdinov.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/KapJI/capital-gains-calculator',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Functions to work with HMRC transaction log."""
from dataclasses import astuple
import datetime
from decimal import Decimal

from .model import HmrcTransactionData, HmrcTransactionLog


def has_key(
    transactions: HmrcTransactionLog, date_index: datetime.date, symbol: str
) -> bool:
    """Check if transaction log has entry for date_index and symbol."""
    return date_index in transactions and symbol in transactions[date_index]


def add_to_list(
    current_list: HmrcTransactionLog,
    date_index: datetime.date,
    symbol: str,
    quantity: Decimal,
    amount: Decimal,
    fees: Decimal,
) -> None:
    """Add entry to given transaction log."""
    if date_index not in current_list:
        current_list[date_index] = {}
    if symbol not in current_list[date_index]:
        current_list[date_index][symbol] = HmrcTransactionData(
            quantity=Decimal(0), amount=Decimal(0), fees=Decimal(0)
        )
    current_quantity, current_amount, current_fees = astuple(
        current_list[date_index][symbol]
    )
    current_list[date_index][symbol] = HmrcTransactionData(
        quantity=current_quantity + quantity,
        amount=current_amount + amount,
        fees=current_fees + fees,
    )



"""Render PDF report with LaTeX."""
from decimal import Decimal
import os
from pathlib import Path
import subprocess
import tempfile

import jinja2

from .const import PACKAGE_NAME, TEMPLATE_NAME
from .model import CapitalGainsReport
from .util import round_decimal, strip_zeros


def render_calculations(
    report: CapitalGainsReport,
    output_path: Path,
    skip_pdflatex: bool = False,
) -> None:
    """Render PDF report."""
    print("Generate calculations report")
    latex_template_env = jinja2.Environment(
        block_start_string="\\BLOCK{",
        block_end_string="}",
        variable_start_string="\\VAR{",
        variable_end_string="}",
        comment_start_string="\\#{",
        comment_end_string="}",
        line_statement_prefix="%%",
        line_comment_prefix="%#",
        trim_blocks=True,
        autoescape=False,
        loader=jinja2.PackageLoader(PACKAGE_NAME, "resources"),
    )
    template = latex_template_env.get_template(TEMPLATE_NAME)
    output_text = template.render(
        report=report,
        round_decimal=round_decimal,
        strip_zeros=strip_zeros,
        Decimal=Decimal,
    )
    generated_file_fd, generated_file = tempfile.mkstemp(suffix=".tex")
    os.write(generated_file_fd, output_text.encode())
    os.close(generated_file_fd)

    # In case of testing
    if skip_pdflatex:
        return
    current_directory = Path.cwd()
    output_filename = "calculations"
    subprocess.run(
        [
            "pdflatex",
            f"-output-directory={current_directory}",
            f"-jobname={output_filename}",
            "-interaction=batchmode",
            generated_file,
        ],
        check=True,
        stdout=subprocess.DEVNULL,
    )
    Path(generated_file).unlink()
    Path(f"{output_filename}.log").unlink()
    Path(f"{output_filename}.aux").unlink()
    Path(f"{output_filename}.pdf").replace(output_path)



"""Parse input files."""
from __future__ import annotations

import csv
import datetime
from decimal import Decimal
import importlib.resources
import operator
from pathlib import Path

from cgt_calc.const import DEFAULT_INITIAL_PRICES_FILE
from cgt_calc.exceptions import UnexpectedColumnCountError
from cgt_calc.model import BrokerTransaction
from cgt_calc.resources import RESOURCES_PACKAGE

from .mssb import read_mssb_transactions
from .schwab import read_schwab_transactions
from .sharesight import read_sharesight_transactions
from .trading212 import read_trading212_transactions


class InitialPricesEntry:
    """Entry from initial stock prices file."""

    def __init__(self, row: list[str], file: str):
        """Create entry from CSV row."""
        if len(row) != 3:
            raise UnexpectedColumnCountError(row, 3, file)
        # date,symbol,price
        self.date = self._parse_date(row[0])
        self.symbol = row[1]
        self.price = Decimal(row[2])

    @staticmethod
    def _parse_date(date_str: str) -> datetime.date:
        """Parse date from string."""
        return datetime.datetime.strptime(date_str, "%b %d, %Y").date()

    def __str__(self) -> str:
        """Return string representation."""
        return f"date: {self.date}, symbol: {self.symbol}, price: {self.price}"


def read_broker_transactions(
    schwab_transactions_file: str | None,
    schwab_awards_transactions_file: str | None,
    trading212_transactions_folder: str | None,
    mssb_transactions_folder: str | None,
    sharesight_transactions_folder: str | None,
) -> list[BrokerTransaction]:
    """Read transactions for all brokers."""
    transactions = []
    if schwab_transactions_file is not None:
        transactions += read_schwab_transactions(
            schwab_transactions_file, schwab_awards_transactions_file
        )
    else:
        print("WARNING: No schwab file provided")

    if trading212_transactions_folder is not None:
        transactions += read_trading212_transactions(trading212_transactions_folder)
    else:
        print("WARNING: No trading212 folder provided")

    if mssb_transactions_folder is not None:
        transactions += read_mssb_transactions(mssb_transactions_folder)
    else:
        print("WARNING: No mssb folder provided")

    if sharesight_transactions_folder is not None:
        transactions += read_sharesight_transactions(sharesight_transactions_folder)
    else:
        print("WARNING: No sharesight file provided")

    transactions.sort(key=operator.attrgetter("date"))
    return transactions


def read_initial_prices(
    initial_prices_file: str | None,
) -> dict[datetime.date, dict[str, Decimal]]:
    """Read initial stock prices from CSV file."""
    initial_prices: dict[datetime.date, dict[str, Decimal]] = {}
    if initial_prices_file is None:
        csv_file = importlib.resources.open_text(
            RESOURCES_PACKAGE, DEFAULT_INITIAL_PRICES_FILE
        )
        lines = list(csv.reader(csv_file))
        csv_file.close()
    else:
        with Path(initial_prices_file).open(encoding="utf-8") as csv_file:
            lines = list(csv.reader(csv_file))
    lines = lines[1:]
    for row in lines:
        entry = InitialPricesEntry(row, initial_prices_file or "default")
        date_index = entry.date
        if date_index not in initial_prices:
            initial_prices[date_index] = {}
        initial_prices[date_index][entry.symbol] = entry.price
    return initial_prices



"""Charles Schwab parser."""
from __future__ import annotations

from collections import defaultdict
import csv
from dataclasses import dataclass
import datetime
from decimal import Decimal
import itertools
from pathlib import Path

from cgt_calc.exceptions import (
    ParsingError,
    SymbolMissingError,
    UnexpectedColumnCountError,
    UnexpectedRowCountError,
)
from cgt_calc.model import ActionType, BrokerTransaction


@dataclass
class AwardPrices:
    """Class to store initial stock prices."""

    award_prices: dict[datetime.date, dict[str, Decimal]]

    def get(self, date: datetime.date, symbol: str) -> Decimal:
        """Get initial stock price at given date."""
        # Award dates may go back for few days, depending on
        # holidays or weekends, so we do a linear search
        # in the past to find the award price
        for i in range(7):
            to_search = date - datetime.timedelta(days=i)

            if (
                to_search in self.award_prices
                and symbol in self.award_prices[to_search]
            ):
                return self.award_prices[to_search][symbol]
        raise Exception(f"Award price is not found for symbol {symbol} for date {date}")


def action_from_str(label: str) -> ActionType:
    """Convert string label to ActionType."""
    if label == "Buy":
        return ActionType.BUY

    if label == "Sell":
        return ActionType.SELL

    if label in [
        "MoneyLink Transfer",
        "Misc Cash Entry",
        "Service Fee",
        "Wire Funds",
        "Funds Received",
        "Journal",
        "Cash In Lieu",
    ]:
        return ActionType.TRANSFER

    if label == "Stock Plan Activity":
        return ActionType.STOCK_ACTIVITY

    if label in ["Qualified Dividend", "Cash Dividend"]:
        return ActionType.DIVIDEND

    if label in ["NRA Tax Adj", "NRA Withholding", "Foreign Tax Paid"]:
        return ActionType.TAX

    if label == "ADR Mgmt Fee":
        return ActionType.FEE

    if label in ["Adjustment", "IRS Withhold Adj"]:
        return ActionType.ADJUSTMENT

    if label in ["Short Term Cap Gain", "Long Term Cap Gain"]:
        return ActionType.CAPITAL_GAIN

    if label == "Spin-off":
        return ActionType.SPIN_OFF

    if label == "Credit Interest":
        return ActionType.INTEREST

    if label == "Reinvest Shares":
        return ActionType.REINVEST_SHARES

    if label == "Reinvest Dividend":
        return ActionType.REINVEST_DIVIDENDS

    if label == "Wire Funds Received":
        return ActionType.WIRE_FUNDS_RECEIVED

    raise ParsingError("schwab transactions", f"Unknown action: {label}")


class SchwabTransaction(BrokerTransaction):
    """Represent single Schwab transaction."""

    def __init__(
        self,
        row: list[str],
        file: str,
    ):
        """Create transaction from CSV row."""
        if len(row) != 9:
            raise UnexpectedColumnCountError(row, 9, file)
        if row[8] != "":
            raise ParsingError(file, "Column 9 should be empty")
        as_of_str = " as of "
        if as_of_str in row[0]:
            index = row[0].find(as_of_str) + len(as_of_str)
            date_str = row[0][index:]
        else:
            date_str = row[0]
        date = datetime.datetime.strptime(date_str, "%m/%d/%Y").date()
        self.raw_action = row[1]
        action = action_from_str(self.raw_action)
        symbol = row[2] if row[2] != "" else None
        description = row[3]
        quantity = Decimal(row[4]) if row[4] != "" else None
        price = Decimal(row[5].replace("$", "")) if row[5] != "" else None
        fees = Decimal(row[6].replace("$", "")) if row[6] != "" else Decimal(0)
        amount = Decimal(row[7].replace("$", "")) if row[7] != "" else None

        currency = "USD"
        broker = "Charles Schwab"
        super().__init__(
            date,
            action,
            symbol,
            description,
            quantity,
            price,
            fees,
            amount,
            currency,
            broker,
        )

    @staticmethod
    def create(
        row: list[str], file: str, awards_prices: AwardPrices
    ) -> SchwabTransaction:
        """Create and post process a SchwabTransaction."""
        transaction = SchwabTransaction(row, file)
        if (
            transaction.price is None
            and transaction.action == ActionType.STOCK_ACTIVITY
        ):
            symbol = transaction.symbol
            if symbol is None:
                raise SymbolMissingError(transaction)
            transaction.price = awards_prices.get(transaction.date, symbol)
        return transaction


def read_schwab_transactions(
    transactions_file: str, schwab_award_transactions_file: str | None
) -> list[BrokerTransaction]:
    """Read Schwab transactions from file."""
    awards_prices = _read_schwab_awards(schwab_award_transactions_file)
    try:
        with Path(transactions_file).open(encoding="utf-8") as csv_file:
            lines = list(csv.reader(csv_file))

            if "Transactions" not in lines[0][0]:
                raise ParsingError(
                    transactions_file,
                    "First line of Schwab transactions file must be something like "
                    "'Transactions for account ...'",
                )

            if len(lines[1]) != 9:
                raise ParsingError(
                    transactions_file,
                    "Second line of Schwab transactions file must be a header"
                    " with 9 columns",
                )

            if "Total" not in lines[-1][0]:
                raise ParsingError(
                    transactions_file,
                    "Last line of Schwab transactions file must be total",
                )

            # Remove headers and footer
            lines = lines[2:-1]
            transactions = [
                SchwabTransaction.create(row, transactions_file, awards_prices)
                for row in lines
            ]
            transactions.reverse()
            return list(transactions)
    except FileNotFoundError:
        print(f"WARNING: Couldn't locate Schwab transactions file({transactions_file})")
        return []


def _read_schwab_awards(
    schwab_award_transactions_file: str | None,
) -> AwardPrices:
    """Read initial stock prices from CSV file."""
    initial_prices: dict[datetime.date, dict[str, Decimal]] = defaultdict(dict)

    lines = []
    if schwab_award_transactions_file is not None:
        try:
            with Path(schwab_award_transactions_file).open(
                encoding="utf-8"
            ) as csv_file:
                lines = list(csv.reader(csv_file))
                # Remove headers
                lines = lines[2:]
        except FileNotFoundError:
            print(
                "WARNING: Couldn't locate Schwab award "
                f"file({schwab_award_transactions_file})"
            )
    else:
        print("WARNING: No schwab award file provided")

    modulo = len(lines) % 3
    if modulo != 0:
        raise UnexpectedRowCountError(
            len(lines) - modulo + 3, schwab_award_transactions_file or ""
        )

    for row in zip(lines[::3], lines[1::3], lines[2::3]):
        if len(row) != 3:
            raise UnexpectedColumnCountError(
                list(itertools.chain(*row)), 3, schwab_award_transactions_file or ""
            )

        lapse_main, _, lapse_data = row

        if len(lapse_main) != 8:
            raise UnexpectedColumnCountError(
                lapse_main, 8, schwab_award_transactions_file or ""
            )
        if len(lapse_data) != 8:
            raise UnexpectedColumnCountError(
                lapse_data, 7, schwab_award_transactions_file or ""
            )

        date_str = lapse_main[0]
        date = datetime.datetime.strptime(date_str, "%Y/%m/%d").date()
        symbol = lapse_main[2] if lapse_main[2] != "" else None
        price = Decimal(lapse_data[3].replace("$", "")) if lapse_data[3] != "" else None
        if symbol is not None and price is not None:
            initial_prices[date][symbol] = price
    return AwardPrices(award_prices=dict(initial_prices))
